/*
 * This software was developed at the National Institute of Standards and Technology
 * by employees of the Federal Government in the course of their official duties.
 * Pursuant to Title 17 Section 105 of the United States Code this software is not
 * subject to copyright protection within the United States and is in the public domain.
 * The software is an experimental system. NIST assumes no responsibility whatsoever for
 * its use by any party.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS." YOU ACCEPT THE SOFTWARE "AS IS" AND ACKNOWLEDGE
 * THAT NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, INCLUDING,
 * WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY. NIST DOES NOT REPRESENT OR WARRANT THAT
 * THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT ANY
 * DEFECTS WILL BE CORRECTED. NIST DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS
 * REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED
 * TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.
 * 
 * You may freely use, copy or distribute this software.Please include appropriate
 * attribution.
 */

package com.simfill.gui;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Vector;

public class LogModel extends java.util.Observable
{
	public LogModel()
	{
		m_namespaces = new HashMap  < String, LogLineAttributes >();
		m_severity   = new Vector   < LogSeverity >();
		m_events     = new ArrayList< LogLine >();
		m_cache      = new StringBuilder();
	}

	public void registerSeverity( String label, int level ) throws Exception
	{
		// Check if an element with the same label or level has already been registered
		for( LogSeverity ls : m_severity )
		{
			if( ls.getSeverity() == level )
			{
				throw new Exception( "LogModel: severity level " + level + " is already registered." );
			}
			
			if( ls.toString().equals( label ) )
			{
				throw new Exception( "LogModel: a severity level with the label '" + label + "' is already registered." );
			}
		}

		m_severity.add( new LogSeverity( label, level ) );
	}
	
	public void registerNamespace( String namespace, String color, int severity ) throws Exception
	{
		// Check if the severity has been registered
		for( LogSeverity ls : m_severity )
		{
			if( ls.getSeverity() == severity )
			{
				// If so, check if the namespace has already been registered
				if( ! m_namespaces.containsKey( namespace ) )
				{
					// If not, add the new entry
					m_namespaces.put( namespace, new LogLineAttributes( color, severity ) );
					return;
				}
				else
				{
					throw new Exception( "LogModel: namespace '" + namespace + "' is already registered." );
				}
			}
		}

		// The severity has not been found
		throw new Exception( "LogModel: severity level " + severity + " is not registered." );
	}
	
	public synchronized void log( String namespace, Object o, boolean useColor )
	{
		log( namespace, o.toString(), useColor );
	}
	
	public synchronized void log( String namespace, String message, boolean useColor )
	{
		if( ! m_namespaces.containsKey( namespace ) )
		{
			// FIXME: if the namespace "err" is not registered, an infinite loop happens
			log( "err", "LogModel: namespace '" + namespace + "' is not registered for the message: " + message, true );
		}

		// TODO sanitize the text
		LogLine ll = new LogLine( message, m_namespaces.get( namespace ), useColor );
		m_events.add( ll );
		
		
		if( m_threshold.getSeverity() >= ll.getAttributes().getSeverity() )
		{
			m_cache.append( ll.getHTML() );
		
			setChanged();
			notifyObservers();
		}

		// FIXME
		System.out.println( message );
	}
	
	public synchronized void clear()
	{
		m_events.clear();
		setChanged();
		notifyObservers();
	}
	
	public void setFilter( String severity ) throws Exception
	{
		// Look for the matching registrered severity
		for( LogSeverity ls : m_severity )
		{
			if( ls.toString().equals( severity ) )
			{
				// If found, update the filter's threshold
				m_threshold = ls;
				printLog();
				return;
			}
		}

		// The severity has not been found
		throw new Exception( "LogModel: severity '" + severity + "' is not registered." );
	}
	
	public Vector< ? > getSeverity()
	{
		return m_severity;
	}
	
	public String getLog()
	{
		return m_cache.toString();
	}
	
	public LogSeverity getFilter()
	{
		return m_threshold;
	}

	private void printLog()
	{
		m_cache = new StringBuilder();

		for( LogLine ll : m_events )
		{
			if( m_threshold.getSeverity() >= ll.getAttributes().getSeverity() )
			{
				m_cache.append( ll.getHTML() );
			}
		}

		setChanged();
		notifyObservers();
	}
	
	// Key: namespace's name
	// Value: color and severity
	private HashMap< String, LogLineAttributes > m_namespaces;
	
	// Threshold used to filter events, based on severity
	private LogSeverity m_threshold;
	
	// Array containing all the events
	private ArrayList< LogLine > m_events;
	
	// Vector containing the severity levels
	private Vector< LogSeverity  > m_severity;

	// Cache containing the text to display
	private StringBuilder m_cache;
}
