/*
 * This software was developed at the National Institute of Standards and Technology
 * by employees of the Federal Government in the course of their official duties.
 * Pursuant to Title 17 Section 105 of the United States Code this software is not
 * subject to copyright protection within the United States and is in the public domain.
 * The software is an experimental system. NIST assumes no responsibility whatsoever for
 * its use by any party.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS." YOU ACCEPT THE SOFTWARE "AS IS" AND ACKNOWLEDGE
 * THAT NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, INCLUDING,
 * WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY. NIST DOES NOT REPRESENT OR WARRANT THAT
 * THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT ANY
 * DEFECTS WILL BE CORRECTED. NIST DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS
 * REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED
 * TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.
 * 
 * You may freely use, copy or distribute this software.Please include appropriate
 * attribution.
 */

package com.simfill;

import java.io.File;
import java.util.Stack;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import com.jaccal.CardException;
import com.jaccal.util.NumUtil;
import com.simfill.gsm.AD;
import com.simfill.gsm.Address;
import com.simfill.gsm.Contact;
import com.simfill.gsm.ELP;
import com.simfill.gsm.Extension;
import com.simfill.gsm.FileInfo;
import com.simfill.gsm.ICCID;
import com.simfill.gsm.IMSI;
import com.simfill.gsm.LOCI;
import com.simfill.gsm.LOCIGPRS;
import com.simfill.gsm.LP;
import com.simfill.gsm.PHASE;
import com.simfill.gsm.PLMN;
import com.simfill.gsm.RawData;
import com.simfill.gsm.RawRecord;
import com.simfill.gsm.SMS;
import com.simfill.gsm.SMSDeliver;
import com.simfill.gsm.SMSReplyPath;
import com.simfill.gsm.SMSSubmit;
import com.simfill.gsm.SMSUserData;
import com.simfill.gsm.SMSUserDataHeaderConcat;
import com.simfill.gsm.SMSUserDataHeaderLargeImage;
import com.simfill.gsm.SMSUserDataHeaderSmallImage;
import com.simfill.gsm.SPN;
import com.simfill.gsm.SST;

public class DataXmlHandler extends DefaultHandler
{
	public DataXmlHandler()
	{
		super();

		m_lastRecIndex = 1;
		
		m_state = new Stack< State >();
		m_state.push( State.INITIALIZED );
		
		m_buffer = new StringBuilder();
	}

	public void startElement( String uri, String localName, String qName, Attributes attributes )
	{
		try
		{
			m_buffer.delete( 0, m_buffer.length() );
			
			switch( m_state.peek() )
			{
				case INITIALIZED:
	
					if( localName.equals( "mf" ) )
					{
						try
						{
							LogManager.getInstance().log( "info", "Initializing the PC/SC interface...", true );
							m_pcsc = new PCSC();
						}
						catch( Exception e )
						{
							LogManager.getInstance().log( "err", "PC/SC interface failed to initalize.", true );
							LogManager.getInstance().log( "excp", e, true );
						}
						
						LogManager.getInstance().log( "info", "Card data loaded.", true );
						
						// Authenticate with PIN1 if provided
						final String pin1 = Registry.getInstance().get( "card.pin1" );
						if( pin1 != null ) m_pcsc.authenticate( PCSC.CHV.PIN1, pin1 );
						
						// Authenticate with PIN2 if provided
						final String pin2 = Registry.getInstance().get( "card.pin2" );
						if( pin2 != null ) m_pcsc.authenticate( PCSC.CHV.PIN2, pin2 );
						
						// Authenticate with ADM code, depending on the card manufacturer
						final String cardType = Registry.getInstance().get( "card.type" );
						if( cardType != null )
						{
							if( cardType.equals( "usimera3" ) )
							{
								m_pcsc.authenticateAxalto(
										Registry.getInstance().get( "card.axalto.keyRefId" ),
										Registry.getInstance().get( "card.axalto.key" ) );
							}
							else if( cardType.equals( "gemxplore" ) )
							{
								m_pcsc.authenticate( PCSC.CHV.ADM1, Registry.getInstance().get( "card.gemplus.adm1" ) );
							}
						}
						
						m_state.push( State.MF_SELECTED );
					}
					break;
					
				case RAW_SELECTED:

					if( localName.equals( "transparent" ) )
					{
						m_struct = FileInfo.STRUCT_TRANSPARENT;
						m_state.push( State.RAW_TRANSPARENT );
					}
					else if( localName.equals( "fixed" ) )
					{
						m_struct = FileInfo.STRUCT_LINEARFIXED;
						m_state.push( State.RAW_FIXED );
					}
					else if( localName.equals( "cyclic" ) )
					{
						m_struct = FileInfo.STRUCT_CYCLIC;
						m_state.push( State.RAW_CYCLIC );
					}
					break;
					
				case RAW_TRANSPARENT:
					
					     if( localName.equals( "id"   ) )	m_state.push( State.RAW_PARSING_ID );
					else if( localName.equals( "data" ) )	m_state.push( State.RAW_PARSING_DATA );
					break;

				case RAW_FIXED:
				case RAW_CYCLIC:
					
					     if( localName.equals( "id"     ) )	m_state.push( State.RAW_PARSING_ID );
					else if( localName.equals( "record" ) )	m_state.push( State.RAW_PARSING_RECORD );
					else if( localName.equals( "empty"  ) )	m_state.push( State.EMPTY );
					break;

				case MF_SELECTED:

					if( localName.equals( "ef_raw" ) )
					{
						m_path = new String[]{ PCSC.MF, null };
						m_pathname = "MF/";
						m_state.push( State.RAW_SELECTED );
					}
					else if( localName.equals( "ef_iccid" ) )
					{
						select( "MF/EF_ICCID", new String[]{ PCSC.MF, PCSC.EF_ICCID }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_TRANSPARENT );
						m_iccid = new ICCID( m_pcsc.getFileInfo().getSize() );
						m_state.push( State.ICCID_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_elp" ) )
					{
						select( "MF/EF_ELP", new String[]{ PCSC.MF, PCSC.EF_ELP }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_TRANSPARENT );
						m_extendedLanguagePreference = new ELP( m_pcsc.getFileInfo().getSize() );
						m_state.push( State.ELP_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "df_gsm" ) )
					{
						LogManager.getInstance().log( "info", "<hr><h3>Processing MF/DF_GSM</h3>", true );
						m_state.push( State.DF_GSM_SELECTED );
					}
					else if( localName.equals( "df_telecom" ) )
					{
						LogManager.getInstance().log( "info", "<hr><h3>Processing MF/DF_TELECOM</h3>", true );
						m_state.push( State.DF_TELECOM_SELECTED );
					}
					break;

				case DF_GSM_SELECTED:
					
					if( localName.equals( "ef_raw" ) )
					{
						m_path = new String[]{ PCSC.MF, PCSC.DF_GSM, null };
						m_pathname = "MF/DF_GSM/";
						m_state.push( State.RAW_SELECTED );
					}
					else if( localName.equals( "ef_lp" ) )
					{
						select( "MF/DF_GSM/EF_LP", new String[]{ PCSC.MF, PCSC.DF_GSM, PCSC.EF_LP }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_TRANSPARENT );
						m_languagePreference = new LP( m_pcsc.getFileInfo().getSize() );
						m_state.push( State.LP_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_imsi" ) )
					{
						select( "MF/DF_GSM/EF_IMSI", new String[]{ PCSC.MF, PCSC.DF_GSM, PCSC.EF_IMSI }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_TRANSPARENT );
						m_imsi = new IMSI( m_pcsc.getFileInfo().getSize() );
						m_state.push( State.IMSI_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_plmnsel" ) )
					{
						select( "MF/DF_GSM/EF_PLMNsel", new String[]{ PCSC.MF, PCSC.DF_GSM, PCSC.EF_PLMNsel }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_TRANSPARENT );
						m_plmn = new PLMN( m_pcsc.getFileInfo().getSize() );
						m_state.push( State.PLMN_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_sst" ) )
					{
						select( "MF/DF_GSM/EF_SST", new String[]{ PCSC.MF, PCSC.DF_GSM, PCSC.EF_SST }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_TRANSPARENT );
						m_sst = new SST( m_pcsc.getFileInfo().getSize() );
						m_state.push( State.SST_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_spn" ) )
					{
						select( "MF/DF_GSM/EF_SPN", new String[]{ PCSC.MF, PCSC.DF_GSM, PCSC.EF_SPN }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_TRANSPARENT );
						m_spn = new SPN( m_pcsc.getFileInfo().getSize() );
						m_state.push( State.SPN_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_fplmn" ) )
					{
						select( "MF/DF_GSM/EF_FPLMN", new String[]{ PCSC.MF, PCSC.DF_GSM, PCSC.EF_FPLMN }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_TRANSPARENT );
						m_plmn = new PLMN( m_pcsc.getFileInfo().getSize() );
						m_state.push( State.PLMN_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_loci" ) )
					{
						select( "MF/DF_GSM/EF_LOCI", new String[]{ PCSC.MF, PCSC.DF_GSM, PCSC.EF_LOCI }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_TRANSPARENT );
						m_loci = new LOCI( m_pcsc.getFileInfo().getSize() );
						m_state.push( State.LOCI_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_locigprs" ) )
					{
						select( "MF/DF_GSM/EF_LOCIGPRS", new String[]{ PCSC.MF, PCSC.DF_GSM, PCSC.EF_LOCIGPRS }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_TRANSPARENT );
						m_locigprs = new LOCIGPRS( m_pcsc.getFileInfo().getSize() );
						m_state.push( State.LOCIGPRS_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_ad" ) )
					{
						select( "MF/DF_GSM/EF_AD", new String[]{ PCSC.MF, PCSC.DF_GSM, PCSC.EF_AD }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_TRANSPARENT );
						m_ad = new AD( m_pcsc.getFileInfo().getSize() );
						m_state.push( State.AD_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_phase" ) )
					{
						select( "MF/DF_GSM/EF_PHASE", new String[]{ PCSC.MF, PCSC.DF_GSM, PCSC.EF_PHASE }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_TRANSPARENT );
						m_phase = new PHASE( m_pcsc.getFileInfo().getSize() );
						m_state.push( State.PHASE_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					break;
					
				case ELP_SELECTED:

					     if( localName.equals( "elp"   ) )	m_state.push( State.ELP_PARSING_ELP );
					else if( localName.equals( "empty" ) )	m_state.push( State.EMPTY );
					break;

				case LP_SELECTED:

					     if( localName.equals( "lp"    ) )	m_state.push( State.LP_PARSING_LP );
					else if( localName.equals( "empty" ) )	m_state.push( State.EMPTY );
					break;

				case PLMN_SELECTED:
					
					     if( localName.equals( "plmn"  ) )	m_state.push( State.PLMN_PARSING_PLMN );
					else if( localName.equals( "empty" ) )	m_state.push( State.EMPTY );
					break;
					
				case PLMN_PARSING_PLMN:
					
					     if( localName.equals( "mcc" ) )	m_state.push( State.PLMN_PARSING_MCC );
					else if( localName.equals( "mnc" ) )	m_state.push( State.PLMN_PARSING_MNC );
					break;
					
				case SST_SELECTED:
					
					if( localName.equals( "service" ) )
					{
						m_sstServiceNumber = 0;
						m_sstAllocated = false;
						m_sstActivated = false;
						
						m_state.push( State.SST_PARSING_SERVICE );
					}
					break;
					
				case SST_PARSING_SERVICE:
					
					     if( localName.equals( "number"    ) )	m_state.push( State.SST_PARSING_NUMBER    );
					else if( localName.equals( "allocated" ) )	m_state.push( State.SST_PARSING_ALLOCATED );
					else if( localName.equals( "activated" ) )	m_state.push( State.SST_PARSING_ACTIVATED );
				    break;
					
				case SPN_SELECTED:
					
					     if( localName.equals( "display" ) ) m_state.push( State.SPN_PARSING_DISPLAY );
					else if( localName.equals( "spn"     ) )
					{
						m_state.push( State.SPN_PARSING_SPN );
						
						String enc = attributes.getValue( "enc" );

						     if( enc == null )             m_spn.setNameEncoding( SPN.Encoding.ENC_DEFAULT );
						else if( enc.equals( "default" ) ) m_spn.setNameEncoding( SPN.Encoding.ENC_DEFAULT );
						else if( enc.equals( "ucs2"    ) ) m_spn.setNameEncoding( SPN.Encoding.ENC_UCS2 );
						else
						{
							LogManager.getInstance().log( "warn", "Encoding '" + enc + "' is not supported. Using default encoding instead.", true );
							m_spn.setNameEncoding( SPN.Encoding.ENC_DEFAULT );
						}
					}
					break;

				case LOCI_SELECTED:
					
					     if( localName.equals( "tmsi"   ) )	m_state.push( State.LOCI_PARSING_TMSI );
					else if( localName.equals( "lai"    ) )	m_state.push( State.LOCI_PARSING_LAI);
					else if( localName.equals( "time"   ) )	m_state.push( State.LOCI_PARSING_TIME );
					else if( localName.equals( "status" ) )	m_state.push( State.LOCI_PARSING_STATUS );
					break;
					
				case LOCI_PARSING_LAI:
					
					     if( localName.equals( "mcc" ) )	m_state.push( State.LOCI_PARSING_LAI_MCC );
					else if( localName.equals( "mnc" ) )	m_state.push( State.LOCI_PARSING_LAI_MNC );
					else if( localName.equals( "lac" ) )	m_state.push( State.LOCI_PARSING_LAI_LAC );
					break;
					
				case LOCIGPRS_SELECTED:
					
					     if( localName.equals( "ptmsi"		) )	m_state.push( State.LOCIGPRS_PARSING_PTMSI );
					else if( localName.equals( "signature"	) )	m_state.push( State.LOCIGPRS_PARSING_SIGNATURE);
					else if( localName.equals( "rai"		) )	m_state.push( State.LOCIGPRS_PARSING_RAI );
					else if( localName.equals( "status"		) )	m_state.push( State.LOCIGPRS_PARSING_STATUS );
					break;
					
				case LOCIGPRS_PARSING_RAI:
					
					     if( localName.equals( "mcc" ) )	m_state.push( State.LOCIGPRS_PARSING_RAI_MCC );
					else if( localName.equals( "mnc" ) )	m_state.push( State.LOCIGPRS_PARSING_RAI_MNC );
					else if( localName.equals( "lac" ) )	m_state.push( State.LOCIGPRS_PARSING_RAI_LAC );
					else if( localName.equals( "rac" ) )	m_state.push( State.LOCIGPRS_PARSING_RAI_RAC );
					break;
				
				case AD_SELECTED:
					
					     if( localName.equals( "opmode"  ) )	m_state.push( State.AD_PARSING_OPMODE  );
					else if( localName.equals( "ofm"	 ) )	m_state.push( State.AD_PARSING_OFM );
					else if( localName.equals( "mnclen"  ) )	m_state.push( State.AD_PARSING_MNCLEN  );
					break;

				case DF_TELECOM_SELECTED:

					if( localName.equals( "ef_raw" ) )
					{
						m_path = new String[]{ PCSC.MF, PCSC.DF_TELECOM, null };
						m_pathname = "MF/DF_TELECOM/";
						m_state.push( State.RAW_SELECTED );
					}
					else if( localName.equals( "ef_adn" ) )
					{
						select( "MF/DF_TELECOM/EF_ADN", new String[]{ PCSC.MF, PCSC.DF_TELECOM, PCSC.EF_ADN }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_LINEARFIXED );
						m_state.push( State.CONTACT_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_fdn" ) )
					{
						select( "MF/DF_TELECOM/EF_FDN", new String[]{ PCSC.MF, PCSC.DF_TELECOM, PCSC.EF_FDN }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_LINEARFIXED );
						m_state.push( State.CONTACT_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_msisdn" ) )
					{
						select( "MF/DF_TELECOM/EF_MSISDN", new String[]{ PCSC.MF, PCSC.DF_TELECOM, PCSC.EF_MSISDN }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_LINEARFIXED );
						m_state.push( State.CONTACT_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_lnd" ) )
					{
						select( "MF/DF_TELECOM/EF_LND", new String[]{ PCSC.MF, PCSC.DF_TELECOM, PCSC.EF_LND }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_CYCLIC );
						m_state.push( State.CONTACT_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_ext1" ) )
					{
						select( "MF/DF_TELECOM/EF_EXT1", new String[]{ PCSC.MF, PCSC.DF_TELECOM, PCSC.EF_EXT1 }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_LINEARFIXED );
						m_state.push( State.EXT_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_ext2" ) )
					{
						select( "MF/DF_TELECOM/EF_EXT2", new String[]{ PCSC.MF, PCSC.DF_TELECOM, PCSC.EF_EXT2 }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_LINEARFIXED );
						m_state.push( State.EXT_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_ext3" ) )
					{
						select( "MF/DF_TELECOM/EF_EXT3", new String[]{ PCSC.MF, PCSC.DF_TELECOM, PCSC.EF_EXT3 }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_LINEARFIXED );
						m_state.push( State.EXT_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_ext4" ) )
					{
						select( "MF/DF_TELECOM/EF_EXT4", new String[]{ PCSC.MF, PCSC.DF_TELECOM, PCSC.EF_EXT4 }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_LINEARFIXED );
						m_state.push( State.EXT_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					else if( localName.equals( "ef_sms" ) )
					{
						select( "MF/DF_TELECOM/EF_SMS", new String[]{ PCSC.MF, PCSC.DF_TELECOM, PCSC.EF_SMS }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_LINEARFIXED );
						m_state.push( State.SMS_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
						
						// Reset the whole file by setting all slots empty
						byte [] emptySlot = new byte[ m_pcsc.getFileInfo().getRecSize() ];
						emptySlot[ 0 ] = SMS.STATUS_FREESPACE;
						for( int it = 1; it < m_pcsc.getFileInfo().getRecSize(); ++it )
						{
							emptySlot[ it ] = ( byte )0xff;
						}
						m_pcsc.reset( NumUtil.hex2String( emptySlot ) );
					}
					else if( localName.equals( "ef_sdn" ) )
					{
						select( "MF/DF_TELECOM/EF_SDN", new String[]{ PCSC.MF, PCSC.DF_TELECOM, PCSC.EF_SDN }, FileInfo.FILETYPE_EF, FileInfo.STRUCT_LINEARFIXED );
						m_state.push( State.CONTACT_SELECTED );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					break;
					
				case CONTACT_SELECTED:
					
					if( localName.equals( "contact" ) )
					{
						m_contact = new Contact( m_pcsc.getFileInfo().getRecSize() );
						m_state.push( State.CONTACT_CREATED );
					}
					else if( localName.equals( "empty" ) )
					{
						m_state.push( State.EMPTY );
					}
					break;
					
				case CONTACT_CREATED:

					if( localName.equals( "description" ) )
					{
						m_state.push( State.CONTACT_PARSING_DESC );
						
						String enc = attributes.getValue( "enc" );

						     if( enc == null )             m_contact.setNameEncoding( Contact.Encoding.ENC_DEFAULT );
						else if( enc.equals( "default" ) ) m_contact.setNameEncoding( Contact.Encoding.ENC_DEFAULT );
						else if( enc.equals( "ucs2"    ) ) m_contact.setNameEncoding( Contact.Encoding.ENC_UCS2 );
						else
						{
							LogManager.getInstance().log( "warn", "Encoding '" + enc + "' is not supported. Using default encoding instead.", true );
							m_contact.setNameEncoding( Contact.Encoding.ENC_DEFAULT );
						}
					}
					else if( localName.equals( "address" ) )
					{
						m_address = new Address();
						m_state.push( State.ADR_PARSING_ADDRESS );
					}
					else if( localName.equals( "capability" ) )
					{
						m_state.push( State.CONTACT_PARSING_CAPABILITY );
					}
					else if( localName.equals( "extension" ) )
					{
						m_state.push( State.CONTACT_PARSING_EXTENSION );
					}
					break;
					
				case EXT_SELECTED:
					
					if( localName.equals( "additional" ) )
					{
						m_extension = new Extension( m_pcsc.getFileInfo().getRecSize(), Extension.EXT_TYPE_ADDDATA );
						m_state.push( State.EXT_ADDDATA );
					}
					else if( localName.equals( "empty"  ) )	m_state.push( State.EMPTY );
					break;
					
				case EXT_ADDDATA:
					
					if( localName.equals( "data" ) )
					{
						m_state.push( State.EXT_PARSING_DATA );
					}
					else if( localName.equals( "extension" ) )
					{
						m_state.push( State.EXT_PARSING_EXT );
					}
					break;
					
				case SMS_SELECTED:

					if( localName.equals( "deliver" ) )
					{
						m_sms = new SMSDeliver( m_pcsc.getFileInfo().getRecSize() );
						m_state.push( State.SMS_DELIVER_CREATED );
					}
					else if( localName.equals( "submit" ) )
					{
						m_sms = new SMSSubmit( m_pcsc.getFileInfo().getRecSize() );
						m_state.push( State.SMS_SUBMIT_CREATED );
					}
					else if( localName.equals( "empty" ) )
					{
						m_state.push( State.EMPTY );
					}
					break;
					
				case SMS_DELIVER_CREATED:
					
					     if( localName.equals( "status" ) )	m_state.push( State.SMS_PARSING_STATUS );
					else if( localName.equals( "sca"    ) )	m_state.push( State.SMS_PARSING_SCA );
					else if( localName.equals( "mms"    ) )	m_state.push( State.SMS_PARSING_MMS );
					else if( localName.equals( "sri"    ) )	m_state.push( State.SMS_PARSING_SRI );
					else if( localName.equals( "rp"     ) )	m_state.push( State.SMS_PARSING_RP );
					else if( localName.equals( "oa"     ) )	m_state.push( State.SMS_PARSING_OA );
					else if( localName.equals( "dcs"    ) )	m_state.push( State.SMS_PARSING_DCS );
					else if( localName.equals( "scts"   ) )	m_state.push( State.SMS_PARSING_SCTS );
					else if( localName.equals( "udh"    ) )	m_state.push( State.SMS_PARSING_UDH );
					else if( localName.equals( "ud"     ) )	m_state.push( State.SMS_PARSING_UD );
					break;

				case SMS_SUBMIT_CREATED:
					
					     if( localName.equals( "status" ) )	m_state.push( State.SMS_PARSING_STATUS );
					else if( localName.equals( "sca"    ) )	m_state.push( State.SMS_PARSING_SCA );
					else if( localName.equals( "rd"     ) )	m_state.push( State.SMS_PARSING_RD );
					else if( localName.equals( "srr"    ) )	m_state.push( State.SMS_PARSING_SRR );
					else if( localName.equals( "rp"     ) )	m_state.push( State.SMS_PARSING_RP );
					else if( localName.equals( "mr"     ) ) m_state.push( State.SMS_PARSING_MR );
					else if( localName.equals( "da"     ) )	m_state.push( State.SMS_PARSING_DA );
					else if( localName.equals( "dcs"    ) )	m_state.push( State.SMS_PARSING_DCS );
					else if( localName.equals( "udh"    ) )	m_state.push( State.SMS_PARSING_UDH );
					else if( localName.equals( "ud"     ) )	m_state.push( State.SMS_PARSING_UD );
					break;					

				case SMS_PARSING_SCA:
				case SMS_PARSING_OA:
				case SMS_PARSING_DA:
					
					if( localName.equals( "address" ) )
					{
						m_address = new Address();
						m_state.push( State.ADR_PARSING_ADDRESS );
					}
					break;
					
				case SMS_PARSING_UDH:
					
					if( localName.equals( "concat" ) )
					{
						m_state.push( State.SMS_PARSING_UDH_CONCAT );
						
						m_sms.add2UDH(
							new SMSUserDataHeaderConcat(
								Integer.parseInt( attributes.getValue( "ref" ) ),
								Integer.parseInt( attributes.getValue( "max" ) ),
								Integer.parseInt( attributes.getValue( "seq" ) ) ) );
					}
					else if( localName.equals( "lrgimg" ) )
					{
						m_state.push( State.SMS_PARSING_UDH_LRGIMG );
						
						m_sms.add2UDH(
							new SMSUserDataHeaderLargeImage(
								Integer.parseInt( attributes.getValue( "pos" ) ),
								new File( attributes.getValue( "file" ) ) ) );
					}
					else if( localName.equals( "smlimg" ) )
					{
						m_state.push( State.SMS_PARSING_UDH_SMLIMG );
						
						m_sms.add2UDH(
							new SMSUserDataHeaderSmallImage(
								Integer.parseInt( attributes.getValue( "pos" ) ),
								new File( attributes.getValue( "file" ) ) ) );
					}
					break;
					
				case ADR_PARSING_ADDRESS:
					
					     if( localName.equals( "ton"    ) )	m_state.push( State.ADR_PARSING_TON );
					else if( localName.equals( "npi"    ) )	m_state.push( State.ADR_PARSING_NPI );
					else if( localName.equals( "number" ) )	m_state.push( State.ADR_PARSING_NUMBER );
					break;
			}
		}
		catch( CardException e )
		{
			LogManager.getInstance().log( "err", "State=" + m_state.peek() + " Exception=" + e, true );
		}
		catch( Exception e )
		{
			LogManager.getInstance().log( "excp", "State=" + m_state.peek() + " Exception=" + e, true );
		}
	}

	public void endElement( String uri, String localName, String qName )
	{
		String data = m_buffer.toString();
		
		State state = m_state.peek();

		try
		{
			switch( state )
			{
				case EMPTY:
					
					if( localName.equals( "empty" ) ) 
					{
						m_state.pop();
						m_lastRecIndex++;
						LogManager.getInstance().log( "rprt", "<li>Empty entry added</li>", true );
					}
					break;
					
				case RAW_SELECTED:
					
					if( localName.equals( "ef_raw" ) )
					{
						m_state.pop();
					}
					break;
					
				case RAW_TRANSPARENT:
					
					if( localName.equals( "transparent" ) )
					{
						m_state.pop();
						LogManager.getInstance().log( "rprt", "</ol>", false );
					}
					break;
					
				case RAW_FIXED:
					
					if( localName.equals( "fixed" ) )
					{
						m_state.pop();
						LogManager.getInstance().log( "rprt", "</ol>", false );
					}
					break;

				case RAW_CYCLIC:
					
					if( localName.equals( "cyclic" ) )
					{
						m_state.pop();
						LogManager.getInstance().log( "rprt", "</ol>", false );
					}
					break;

				case RAW_PARSING_ID:
					
					if( localName.equals( "id" ) )
					{
						m_state.pop();
						m_path[ m_path.length - 1 ] = data;
						select( m_pathname + data, m_path, FileInfo.FILETYPE_EF, m_struct );
						LogManager.getInstance().log( "rprt", "<ol>", false );
					}
					break;
					
				case RAW_PARSING_DATA:
					
					if( localName.equals( "data" ) )
					{
						m_state.pop();
						m_pcsc.update( new RawData( data.length() >> 1, data ) );
					}
					break;
					
				case RAW_PARSING_RECORD:
					
					if( localName.equals( "record" ) )
					{
						m_state.pop();
						m_pcsc.update( m_lastRecIndex++, new RawRecord( ( short )( data.length() >> 1 ), data ) );
					}
					break;

 				case MF_SELECTED:
					
					if( localName.equals( "mf" ) )
					{
						m_state.pop();
						m_pcsc.shutdown();
						LogManager.getInstance().log( "info", "<hr>Done.", true );
					}
					break;

				case ELP_SELECTED:
					
					if( localName.equals( "ef_elp" ) )
					{
						m_state.pop();
						m_pcsc.update( m_extendedLanguagePreference );
						LogManager.getInstance().log( "rprt", "</ol>", false );
					}
					break;
					
				case ELP_PARSING_ELP:

					if( localName.equals( "elp" ) )
					{
						m_state.pop();
						m_extendedLanguagePreference.setLang( m_lastRecIndex++, data );
					}
					break;

				case ICCID_SELECTED:
					
					if( localName.equals( "ef_iccid" ) )
					{
						m_state.pop();
						m_iccid.setICCID( data );
						m_pcsc.update( m_iccid );
						LogManager.getInstance().log( "rprt", "</ol>", false );
					}
					break;
					
				case DF_GSM_SELECTED:
					
					if( localName.equals( "df_gsm" ) )
					{
						m_state.pop();
					}
					break;
					
				case LP_SELECTED:
					
					if( localName.equals( "ef_lp" ) )
					{
						m_state.pop();
						m_pcsc.update( m_languagePreference );
						LogManager.getInstance().log( "rprt", "</ol>", false );
					}
					break;
					
				case LP_PARSING_LP:

					if( localName.equals( "lp" ) )
					{
						m_state.pop();

						byte lp;
						if( data.equals( "german"      ) ) lp = LP.LANG_GERMAN     ; else
						if( data.equals( "english"     ) ) lp = LP.LANG_ENGLISH    ; else
						if( data.equals( "italian"     ) ) lp = LP.LANG_ITALIAN    ; else
						if( data.equals( "french"      ) ) lp = LP.LANG_FRENCH     ; else
						if( data.equals( "spanish"     ) ) lp = LP.LANG_SPANISH    ; else
						if( data.equals( "dutch"       ) ) lp = LP.LANG_DUTCH      ; else
						if( data.equals( "swedish"     ) ) lp = LP.LANG_SWEDISH    ; else
						if( data.equals( "danish"      ) ) lp = LP.LANG_DANISH     ; else
						if( data.equals( "portuguese"  ) ) lp = LP.LANG_PORTUGUESE ; else
						if( data.equals( "finnish"     ) ) lp = LP.LANG_FINNISH    ; else
						if( data.equals( "norwegian"   ) ) lp = LP.LANG_NORWEGIAN  ; else
						if( data.equals( "greek"       ) ) lp = LP.LANG_GREEK      ; else
						if( data.equals( "turkish"     ) ) lp = LP.LANG_TURKISH    ; else
						if( data.equals( "hungarian"   ) ) lp = LP.LANG_HUNGARIAN  ; else
						if( data.equals( "polish"      ) ) lp = LP.LANG_POLISH     ; else
						if( data.equals( "unspecified" ) ) lp = LP.LANG_UNSPECIFIED; else
						if( data.equals( "czech"       ) ) lp = LP.LANG_CZECH      ; else
						if( data.equals( "hebrew"      ) ) lp = LP.LANG_HEBREW     ; else
						if( data.equals( "arabic"      ) ) lp = LP.LANG_ARABIC     ; else
						if( data.equals( "russian"     ) ) lp = LP.LANG_RUSSIAN    ; else
						{
							LogManager.getInstance().log( "warn", "Language '" + data + "' is unknown. Using language 'unspecified' instead.", true );
							lp = LP.LANG_UNSPECIFIED;
						}
						m_languagePreference.setLang( m_lastRecIndex++, lp );
					}
					break;
					
				case IMSI_SELECTED:
					
					if( localName.equals( "ef_imsi" ) )
					{
						m_state.pop();
						m_imsi.setIMSI( data );
						m_pcsc.update( m_imsi );
						LogManager.getInstance().log( "rprt", "</ol>", false );
					}
					break;
					
				case PLMN_SELECTED:
					
					if( localName.equals( "ef_plmnsel" )
					 || localName.equals( "ef_fplmn" ) )
					{
						m_state.pop();
						m_pcsc.update( m_plmn );
						LogManager.getInstance().log( "rprt", "</ol>", false );
					}
					break;
					
				case PLMN_PARSING_PLMN:
					
					if( localName.equals( "plmn" ) )
					{
						m_state.pop();
						m_lastRecIndex++;
					}
					break;
	
				case PLMN_PARSING_MCC:
					
					if( localName.equals( "mcc" ) )
					{
						m_state.pop();
						m_plmn.setMCC( 3 * m_lastRecIndex, data );
					}
					break;
					
				case PLMN_PARSING_MNC:
					
					if( localName.equals( "mnc" ) )
					{
						m_state.pop();
						m_plmn.setMNC( 3 * m_lastRecIndex, data );
					}
					break;
					
				case SST_SELECTED:
					
					if( localName.equals( "ef_sst" ) )
					{
						m_state.pop();
						m_pcsc.update( m_sst );
						LogManager.getInstance().log( "rprt", "</ol>", false );
					}
					break;
					
				case SST_PARSING_SERVICE:
					
					if( localName.equals( "service" ) )
					{
						m_state.pop();
						m_sst.setService( m_sstServiceNumber, m_sstAllocated, m_sstActivated );
					}
					break;
					
				case SST_PARSING_NUMBER:
					
					if( localName.equals( "number" ) )
					{
						m_state.pop();
						m_sstServiceNumber = Integer.parseInt( data );
					}
					break;

				case SST_PARSING_ALLOCATED:
					
					if( localName.equals( "allocated" ) )
					{
						m_state.pop();
						m_sstAllocated = Boolean.parseBoolean( data );
					}
					break;

				case SST_PARSING_ACTIVATED:
					
					if( localName.equals( "activated" ) )
					{
						m_state.pop();
						m_sstActivated = Boolean.parseBoolean( data );
					}
					break;

				case SPN_SELECTED:
					
					if( localName.equals( "ef_spn" ) )
					{
						m_state.pop();
						m_pcsc.update( m_spn );
						LogManager.getInstance().log( "rprt", "</ol>", false );
					}
					break;
					
				case SPN_PARSING_DISPLAY:
					
					if( localName.equals( "display" ) )
					{
						m_state.pop();

						SPN.DisplayCondition dcond;
						if( data.equals( "required"     ) ) dcond = SPN.DisplayCondition.REQUIRED    ; else
						if( data.equals( "not required" ) ) dcond = SPN.DisplayCondition.NOT_REQUIRED; else
						{
							dcond = SPN.DisplayCondition.REQUIRED;
							LogManager.getInstance().log( "warn", "Display condition '" + data + "' in SPN is unknown. Using condition 'required' instead.", true );
						}
						m_spn.setDisplayCondition( dcond );
					}
					break;
					
				case SPN_PARSING_SPN:

					if( localName.equals( "spn" ) )
					{
						m_state.pop();
						m_spn.setSPN( data );
					}
					break;

				case LOCI_SELECTED:
					
					if( localName.equals( "ef_loci" ) )
					{
						m_state.pop();
						m_pcsc.update( m_loci );
						LogManager.getInstance().log( "rprt", "</ol>", false );
					}
					break;
					
				case LOCI_PARSING_TMSI:
					
					if( localName.equals( "tmsi" ) )
					{
						m_state.pop();
						m_loci.setTMSI( data );
					}
					break;
					
				case LOCI_PARSING_LAI:
					
					if( localName.equals( "lai" ) )
					{
						m_state.pop();
					}
					break;
					
				case LOCI_PARSING_LAI_MCC:
					
					if( localName.equals( "mcc" ) )
					{
						m_state.pop();
						m_loci.setMCC( data );
					}
					break;
					
				case LOCI_PARSING_LAI_MNC:
					
					if( localName.equals( "mnc" ) )
					{
						m_state.pop();
						m_loci.setMNC( data );
					}
					break;
					
				case LOCI_PARSING_LAI_LAC:
					
					if( localName.equals( "lac" ) )
					{
						m_state.pop();
						m_loci.setLAC( data );
					}
					break;
					
				case LOCI_PARSING_TIME:
					
					if( localName.equals( "time" ) )
					{
						m_state.pop();
						m_loci.setTMSITime( data );
					}
					break;
					
				case LOCI_PARSING_STATUS:
					
					if( localName.equals( "status" ) )
					{
						m_state.pop();

						byte status;
						if( data.equals( "updated" ) )                   status = LOCI.STATUS_UPDATED             ; else
						if( data.equals( "not updated" ) )               status = LOCI.STATUS_NOT_UPDATED         ; else
						if( data.equals( "plmn not allowed" ) )          status = LOCI.STATUS_PLMN_NOT_ALLOWED    ; else
						if( data.equals( "location area not allowed" ) ) status = LOCI.STATUS_LOC_AREA_NOT_ALLOWED; else
						if( data.equals( "reserved" ) )                  status = LOCI.STATUS_RESERVED            ; else
						{
							status = LOCI.STATUS_RESERVED;
							LogManager.getInstance().log( "warn", "Update status '" + data + "' in LOCI in unknown. Using status 'reserved' instead.", true );
						}
						m_loci.setLocationUpdateStatus( status );
					}
					break;

				case LOCIGPRS_SELECTED:
					
					if( localName.equals( "ef_locigprs" ) )
					{
						m_state.pop();
						m_pcsc.update( m_locigprs );
						LogManager.getInstance().log( "rprt", "</ol>", false );
					}
					break;

				case LOCIGPRS_PARSING_PTMSI:
					
					if( localName.equals( "ptmsi" ) )
					{
						m_state.pop();
						m_locigprs.setPTMSI( data );
					}
					break;

				case LOCIGPRS_PARSING_SIGNATURE:
					
					if( localName.equals( "signature" ) )
					{
						m_state.pop();
						m_locigprs.setPTMSISignature( data );
					}
					break;

				case LOCIGPRS_PARSING_RAI:
					
					if( localName.equals( "rai" ) )
					{
						m_state.pop();
					}
					break;
					
				case LOCIGPRS_PARSING_RAI_MCC:
					
					if( localName.equals( "mcc" ) )
					{
						m_state.pop();
						m_locigprs.setMCC( data );
					}
					break;
					
				case LOCIGPRS_PARSING_RAI_MNC:
					
					if( localName.equals( "mnc" ) )
					{
						m_state.pop();
						m_locigprs.setMNC( data );
					}
					break;
					
				case LOCIGPRS_PARSING_RAI_LAC:
					
					if( localName.equals( "lac" ) )
					{
						m_state.pop();
						m_locigprs.setLAC( data );
					}
					break;

				case LOCIGPRS_PARSING_RAI_RAC:
					
					if( localName.equals( "rac" ) )
					{
						m_state.pop();
						m_locigprs.setRAC( data );
					}
					break;

				case LOCIGPRS_PARSING_STATUS:
					
					if( localName.equals( "status" ) )
					{
						m_state.pop();

						byte status;
						if( data.equals( "updated" ) )                  status = LOCIGPRS.STATUS_UPDATED              ; else
						if( data.equals( "not updated" ) )              status = LOCIGPRS.STATUS_NOT_UPDATED          ; else
						if( data.equals( "plmn not allowed" ) )         status = LOCIGPRS.STATUS_PLMN_NOT_ALLOWED     ; else
						if( data.equals( "routing area not allowed" ) ) status = LOCIGPRS.STATUS_ROUT_AREA_NOT_ALLOWED; else
						if( data.equals( "reserved" ) )                 status = LOCIGPRS.STATUS_RESERVED             ; else
						{
							status = LOCIGPRS.STATUS_RESERVED;
							LogManager.getInstance().log( "warn", "Update status '" + data + "' in LOCIGPRS in unknown. Using status 'reserved' instead.", true );
						}
						m_locigprs.setRoutingUpdateStatus( status );
					}
					break;
					
				case AD_SELECTED:
					
					if( localName.equals( "ef_ad" ) )
					{
						m_state.pop();
						m_pcsc.update( m_ad );
						LogManager.getInstance().log( "rprt", "</ol>", false );
					}
					break;
					
				case AD_PARSING_OPMODE:
					
					if( localName.equals( "opmode" ) )
					{
						m_state.pop();
						
						byte opmode = AD.AD_OPMOD_NORMAL;

						     if( data.equals( "normal"					) ) opmode = AD.AD_OPMOD_NORMAL;
						else if( data.equals( "type approval"			) ) opmode = AD.AD_OPMOD_TYPEAPPROVAL;
						else if( data.equals( "normal specific"			) ) opmode = AD.AD_OPMOD_NORMAL       | AD.AD_OPMOD_MASK_SPECIFIC;
						else if( data.equals( "type approval specific"	) ) opmode = AD.AD_OPMOD_TYPEAPPROVAL | AD.AD_OPMOD_MASK_SPECIFIC;
						else if( data.equals( "maintenance"				) ) opmode = AD.AD_OPMOD_MAINTENANCE;
						else if( data.equals( "celltest"				) ) opmode = AD.AD_OPMOD_CELLTEST;

						m_ad.setOpMode( opmode );
					}
					break;
					
				case AD_PARSING_OFM:
					
					if( localName.equals( "ofm" ) )
					{
						m_state.pop();
						
						short ofm = data.equals( "activated" ) ? AD.AD_INFO_OFM_ACTIVATED : AD.AD_INFO_OFM_DISABLED;
						
						m_ad.setAddInfo( ( short )( m_ad.getAddInfo() | ofm ) );
					}
					break;
					
				case AD_PARSING_MNCLEN:
					
					if( localName.equals( "mnclen" ) )
					{
						m_state.pop();
						m_ad.setMNCLength( Byte.parseByte( data ) );
					}
					break;

				case PHASE_SELECTED:
					
					if( localName.equals( "ef_phase" ) )
					{
						m_state.pop();
						m_phase.setPhase( NumUtil.toStringHex( data )[ 0 ] );
						m_pcsc.update( m_phase );
						LogManager.getInstance().log( "rprt", "</ol>", false );
					}
					break;
					
				case DF_TELECOM_SELECTED:
	
					if( localName.equals( "df_telecom" ) )
					{
						m_state.pop();
					}
					break;
					
				case CONTACT_SELECTED:
	
					if( localName.equals( "ef_adn" )
					 || localName.equals( "ef_fdn" )
					 || localName.equals( "ef_msisdn" )
					 || localName.equals( "ef_lnd" ) )
					{
						m_state.pop();
						LogManager.getInstance().log( "rprt", "</ol>", false );
					}
					break;
	
				case CONTACT_CREATED:
					
					if( localName.equals( "contact" ) )
					{
						m_state.pop();
						m_contact.setAddress( m_address );
						m_pcsc.update( m_lastRecIndex++, m_contact );
					}
					break;
	
				case CONTACT_PARSING_DESC:
					
					if( localName.equals( "description" ) )
					{
						m_state.pop();
						m_contact.setName( data );
					}
					break;

				case CONTACT_PARSING_CAPABILITY:
					
					if( localName.equals( "capability" ) )
					{
						m_state.pop();
						m_contact.setCapability( ( byte )Integer.parseInt( data ) );
					}
					break;
					
				case CONTACT_PARSING_EXTENSION:
					
					if( localName.equals( "extension" ) )
					{
						m_state.pop();
						m_contact.setExtension( Byte.parseByte( data ) );
					}
					break;
					
				case EXT_SELECTED:
					
					if( localName.equals( "ef_ext1" )
					 || localName.equals( "ef_ext2" )
					 || localName.equals( "ef_ext3" )
					 || localName.equals( "ef_ext4" ) )
					{
						m_state.pop();
						LogManager.getInstance().log( "rprt", "</ol>", false );
					}
					break;
					
				case EXT_ADDDATA:
					
					if( localName.equals( "additional" ) )
					{
						m_state.pop();
						m_pcsc.update( m_lastRecIndex++, m_extension );
					}
					break;
					
				case EXT_PARSING_DATA:
					
					if( localName.equals( "data" ) )
					{
						m_state.pop();
						m_extension.setData( data );
					}
					break;
					
				case EXT_PARSING_EXT:
					
					if( localName.equals( "extension" ) )
					{
						m_state.pop();
						m_extension.setExtension( Byte.valueOf( data ) );
					}
					break;

				case SMS_SELECTED:
					
					if( localName.equals( "ef_sms" ) )
					{
						m_state.pop();
						LogManager.getInstance().log( "rprt", "</ol>", false );
					}
					break;
					
				case SMS_DELIVER_CREATED:
					
					if( localName.equals( "deliver" ) )
					{
						m_state.pop();
						m_pcsc.update( m_lastRecIndex++, m_sms );
					}
					break;

				case SMS_SUBMIT_CREATED:
					
					if( localName.equals( "submit" ) )
					{
						m_state.pop();
						m_pcsc.update( m_lastRecIndex++, m_sms );
					}
					break;

				case SMS_PARSING_STATUS:
					
					if( localName.equals( "status" ) )
					{
						m_state.pop();
						
						byte status;
						if( data.equals( "free"     ) ) status = SMS.STATUS_FREESPACE   ; else
						if( data.equals( "deleted"  ) ) status = SMS.STATUS_FREESPACE   ; else
						if( data.equals( "read"     ) ) status = SMS.STATUS_IN_READ     ; else
						if( data.equals( "not read" ) ) status = SMS.STATUS_IN_NOT_READ ; else
						if( data.equals( "sent"     ) ) status = SMS.STATUS_OUT_SENT    ; else
						if( data.equals( "not sent" ) ) status = SMS.STATUS_OUT_NOT_SENT; else
						{
							LogManager.getInstance().log( "warn", "SMS status '" + data + "' unknown. Using status 'freespace / deleted' instead.", true );
							status = SMS.STATUS_FREESPACE; // FIXME
						}
						m_sms.setStatus( status );
					}
					break;
					
				case SMS_PARSING_SCA:
					
					if( localName.equals( "sca" ) )
					{
						m_state.pop();
						m_sms.setSCA( m_address );
					}
					break;
					
				case SMS_PARSING_MMS:
					
					if( localName.equals( "mms" ) )
					{
						m_state.pop();
						( ( SMSDeliver )m_sms ).setMMS( data.equals( "true" ) || data.equals( "1" ) ? SMSDeliver.MMS_MORE_MESSAGES : SMSDeliver.MMS_NO_MORE_MESSAGES );
					}
					break;
					
				case SMS_PARSING_RD:
					
					if( localName.equals( "rd" ) )
					{
						m_state.pop();
						( ( SMSSubmit )m_sms ).setRD( data.equals( "true" ) || data.equals( "1" ) ? SMSSubmit.RD_REJECT : SMSSubmit.RD_ACCEPT );
					}
					break;
					
				case SMS_PARSING_SRI:
					
					if( localName.equals( "sri" ) )
					{
						m_state.pop();
						( ( SMSDeliver )m_sms ).setSRI( data.equals( "true" ) || data.equals( "1" ) ? SMSDeliver.SRI_DO_REPORT : SMSDeliver.SRI_DONT_REPORT );
					}
					break;
					
				case SMS_PARSING_SRR:
					
					if( localName.equals( "srr" ) )
					{
						m_state.pop();
						( ( SMSSubmit )m_sms ).setSRR( data.equals( "true" ) || data.equals( "1" ) ? SMSSubmit.SRR_DO_REPORT : SMSSubmit.SRR_DONT_REPORT );
					}
					break;

				case SMS_PARSING_RP:
					
					if( localName.equals( "rp" ) )
					{
						m_state.pop();
						( ( SMSReplyPath )m_sms ).setRP( data.equals( "true" ) || data.equals( "1" ) ? SMSReplyPath.RP_SET : SMSReplyPath.RP_NOT_SET );
					}
					break;

				case SMS_PARSING_MR:
					
					if( localName.equals( "mr" ) )
					{
						m_state.pop();
						( ( SMSSubmit )m_sms ).setMR( Short.parseShort( data ) );
					}
					break;
					
				case SMS_PARSING_OA:
					
					if( localName.equals( "oa" ) )
					{
						m_state.pop();
						( ( SMSDeliver )m_sms ).setOA( m_address );
					}
					break;
					
				case SMS_PARSING_DA:
					
					if( localName.equals( "da" ) )
					{
						m_state.pop();
						( ( SMSSubmit )m_sms ).setDA( m_address );
					}
					break;
					
				case SMS_PARSING_DCS:
					
					if( localName.equals( "dcs" ) )
					{
						m_state.pop();
						
						byte dcs;
						if( data.equals( "default" ) ) dcs = SMSUserData.DCS_7BITS_DEFAULT; else
						if( data.equals( "ucs2"    ) ) dcs = SMSUserData.DCS_16BITS_UCS2  ; else
						{
							LogManager.getInstance().log( "warn", "SMS encoding '" + data + "' unknown. Using default encoding instead.", true );
							dcs = SMSUserData.DCS_7BITS_DEFAULT;
						}
						( ( SMSUserData )m_sms ).setDCS( dcs );
					}
					break;
					
				case SMS_PARSING_SCTS:
					
					if( localName.equals( "scts" ) )
					{
						m_state.pop();

						( ( SMSDeliver )m_sms ).setSCTS( data );
					}
					break;

				case SMS_PARSING_UDH:
					
					if( localName.equals( "udh" ) )
					{
						m_state.pop();
					}
					break;
					
				case SMS_PARSING_UDH_CONCAT:
					
					if( localName.equals( "concat" ) )
					{
						m_state.pop();
					}
					break;
					
				case SMS_PARSING_UDH_LRGIMG:
					
					if( localName.equals( "lrgimg" ) )
					{
						m_state.pop();
					}
					break;
					
				case SMS_PARSING_UDH_SMLIMG:
					
					if( localName.equals( "smlimg" ) )
					{
						m_state.pop();
					}
					break;

				case SMS_PARSING_UD:
					
					if( localName.equals( "ud" ) )
					{
						m_state.pop();
						if( data != null && data.length() > 0 ) 
						{
							( ( SMSUserData )m_sms ).setUD( data );
						}
					}
					break;
			
				case ADR_PARSING_ADDRESS:
				
					if( localName.equals( "address" ) )
					{
						m_state.pop();
					}
					break;
			
				case ADR_PARSING_TON:
					
					if( localName.equals( "ton" ) )
					{
						m_state.pop();
						
						byte ton;
						if( data.equals( "unknown"          ) ) ton = Address.TON_UNKNOWN         ; else
						if( data.equals( "international"    ) ) ton = Address.TON_INTERNATIONAL   ; else
						if( data.equals( "national"         ) ) ton = Address.TON_NATIONAL        ; else
						if( data.equals( "network specific" ) ) ton = Address.TON_NETWORK_SPECIFIC; else
						if( data.equals( "subscriber"       ) ) ton = Address.TON_SUBSCRIBER      ; else
						if( data.equals( "alphanumeric"     ) ) ton = Address.TON_ALPHANUMERIC    ; else
						if( data.equals( "abbreviated"      ) ) ton = Address.TON_ABBREVIATED     ; else
						if( data.equals( "reserved"         ) ) ton = Address.TON_RESERVED        ; else
						{
							LogManager.getInstance().log( "warn", "TON '" + data + "' unknown. Using 'unknown' instead.", true );
							ton = Address.TON_UNKNOWN;
						}
						m_address.setTON( ton );
					}
					break;

				case ADR_PARSING_NPI:
					
					if( localName.equals( "npi" ) )
					{
						m_state.pop();
						
						byte npi;
						if( data.equals( "unknown"                   ) ) npi = Address.NPI_UNKNOWN                  ; else
						if( data.equals( "telephone"                 ) ) npi = Address.NPI_TELEPHONE                ; else
						if( data.equals( "data"                      ) ) npi = Address.NPI_DATA                     ; else
						if( data.equals( "telex"                     ) ) npi = Address.NPI_TELEX                    ; else
						if( data.equals( "service center specific 1" ) ) npi = Address.NPI_SERVICE_CENTER_SPECIFIC_1; else
						if( data.equals( "service center specific 2" ) ) npi = Address.NPI_SERVICE_CENTER_SPECIFIC_2; else
						if( data.equals( "national"                  ) ) npi = Address.NPI_NATIONAL                 ; else
						if( data.equals( "private"                   ) ) npi = Address.NPI_TELEX                    ; else
						if( data.equals( "ermes"                     ) ) npi = Address.NPI_ERMES                    ; else
						if( data.equals( "reserved"                  ) ) npi = Address.NPI_RESERVED                 ; else
						{
							LogManager.getInstance().log( "warn", "NPI '" + data + "' unknown. Using 'unknown' instead.", true );
							npi = Address.NPI_UNKNOWN;
						}
						m_address.setNPI( npi );
					}
					break;

				case ADR_PARSING_NUMBER:
					
					if( localName.equals( "number" ) )
					{
						m_state.pop();
						m_address.setNumber( data );
					}
					break;					
			}	
		}
		catch( CardException e )
		{
			LogManager.getInstance().log( "err", "State=" + state + " Exception=" + e, true );
		}
		catch( Exception e )
		{
			LogManager.getInstance().log( "excp", "State=" + state + " Exception=" + e, true );
		}
	}

	public void characters( char[] ch, int start, int length ) throws SAXException
	{
		switch( m_state.peek() )
		{
			case RAW_PARSING_ID:
			case RAW_PARSING_DATA:
			case RAW_PARSING_RECORD:
			case ELP_PARSING_ELP:
			case ICCID_SELECTED:
			case IMSI_SELECTED:
			case CONTACT_PARSING_DESC:
			case CONTACT_PARSING_CAPABILITY:
			case CONTACT_PARSING_EXTENSION:
			case EXT_PARSING_DATA:
			case EXT_PARSING_EXT:
			case SMS_PARSING_STATUS:
			case SMS_PARSING_MMS:
			case SMS_PARSING_RP:
			case SMS_PARSING_SRI:
			case SMS_PARSING_DCS:
			case SMS_PARSING_SCTS:
			case SMS_PARSING_UDH:
			case SMS_PARSING_UDH_CONCAT:
			case SMS_PARSING_UDH_LRGIMG:
			case SMS_PARSING_UDH_SMLIMG:
			case SMS_PARSING_UD:
			case SMS_PARSING_RD:
			case SMS_PARSING_SRR:
			case SMS_PARSING_MR:
			case SMS_PARSING_DA:
			case LP_PARSING_LP:	
			case PLMN_PARSING_MCC:
			case PLMN_PARSING_MNC:
			case SST_PARSING_NUMBER:
			case SST_PARSING_ALLOCATED:
			case SST_PARSING_ACTIVATED:
			case SPN_PARSING_DISPLAY:
			case SPN_PARSING_SPN:
			case LOCI_PARSING_TMSI:
			case LOCI_PARSING_LAI_MCC:
			case LOCI_PARSING_LAI_MNC:
			case LOCI_PARSING_LAI_LAC:
			case LOCI_PARSING_TIME:
			case LOCI_PARSING_STATUS:
			case LOCIGPRS_PARSING_PTMSI:
			case LOCIGPRS_PARSING_SIGNATURE:
			case LOCIGPRS_PARSING_RAI_MCC:
			case LOCIGPRS_PARSING_RAI_MNC:
			case LOCIGPRS_PARSING_RAI_LAC:
			case LOCIGPRS_PARSING_RAI_RAC:
			case LOCIGPRS_PARSING_STATUS:
			case AD_PARSING_OPMODE:
			case AD_PARSING_OFM:
			case AD_PARSING_MNCLEN:
			case PHASE_SELECTED:
			case ADR_PARSING_TON:
			case ADR_PARSING_NPI:
			case ADR_PARSING_NUMBER:

				m_buffer.append( ch, start, length );
				break;
		}
	}
	
	private void select( String filename, String[] path, byte type, byte struct ) throws Exception
	{
		LogManager.getInstance().log( "info", "<hr>Processing " + filename, true );

		m_pcsc.select( path );
		if( m_pcsc.getFileInfo().getType() != type || m_pcsc.getFileInfo().getStruct() != struct )
		{
			throw new Exception( "Warning: error selecting " + filename + ": the file won't be populated" );
		}
		
		switch( struct )
		{
			case FileInfo.STRUCT_TRANSPARENT:
			
				m_lastRecIndex = 0;
				break;
				
			case FileInfo.STRUCT_LINEARFIXED:
			case FileInfo.STRUCT_CYCLIC:
				
				m_lastRecIndex = 1;
				m_pcsc.reset();
				break;
				
			default:

				throw new Exception( "Warning: select: invalid file structure requested." );
		}
	}

	// Instance of the PC/SC wrapper
	private PCSC m_pcsc;
	
	// States of the parser
	private enum State
	{
		INITIALIZED,
		EMPTY,
		
		RAW_SELECTED,
		RAW_TRANSPARENT,
		RAW_FIXED,
		RAW_CYCLIC,
		RAW_PARSING_ID,
		RAW_PARSING_DATA,
		RAW_PARSING_RECORD,
		
		MF_SELECTED,
		DF_TELECOM_SELECTED,
		DF_GSM_SELECTED,
		ELP_SELECTED,
		ELP_PARSING_ELP,
		ICCID_SELECTED,
	
		CONTACT_SELECTED,
		CONTACT_CREATED,
		CONTACT_PARSING_DESC,
		CONTACT_PARSING_CAPABILITY,
		CONTACT_PARSING_EXTENSION,
	
		LP_SELECTED,
		LP_PARSING_LP,
		
		IMSI_SELECTED,
		
		PLMN_SELECTED,
		PLMN_PARSING_PLMN,
		PLMN_PARSING_MCC,
		PLMN_PARSING_MNC,
		
		SST_SELECTED,
		SST_PARSING_SERVICE,
		SST_PARSING_NUMBER,
		SST_PARSING_ALLOCATED,
		SST_PARSING_ACTIVATED,
		
		SPN_SELECTED,
		SPN_PARSING_DISPLAY,
		SPN_PARSING_SPN,
		
		LOCI_SELECTED,
		LOCI_PARSING_TMSI,
		LOCI_PARSING_LAI,
		LOCI_PARSING_LAI_MCC,
		LOCI_PARSING_LAI_MNC,
		LOCI_PARSING_LAI_LAC,
		LOCI_PARSING_TIME,
		LOCI_PARSING_STATUS,
		
		LOCIGPRS_SELECTED,
		LOCIGPRS_PARSING_PTMSI,
		LOCIGPRS_PARSING_SIGNATURE,
		LOCIGPRS_PARSING_RAI,
		LOCIGPRS_PARSING_RAI_MCC,
		LOCIGPRS_PARSING_RAI_MNC,
		LOCIGPRS_PARSING_RAI_LAC,
		LOCIGPRS_PARSING_RAI_RAC,
		LOCIGPRS_PARSING_STATUS,
		
		AD_SELECTED,
		AD_PARSING_OPMODE,
		AD_PARSING_OFM,
		AD_PARSING_MNCLEN,
		
		PHASE_SELECTED,
	
		ADR_PARSING_ADDRESS,
		ADR_PARSING_TON,
		ADR_PARSING_NPI,
		ADR_PARSING_NUMBER,

		EXT_SELECTED,
		EXT_ADDDATA,
		EXT_PARSING_DATA,
		EXT_PARSING_EXT,

		SMS_SELECTED,
		SMS_DELIVER_CREATED,
		SMS_SUBMIT_CREATED,
		SMS_PARSING_STATUS,
		SMS_PARSING_SCA,
		SMS_PARSING_MMS,
		SMS_PARSING_RP,
		SMS_PARSING_SRI,
		SMS_PARSING_OA,
		SMS_PARSING_DCS,
		SMS_PARSING_SCTS,
		SMS_PARSING_UDH,
		SMS_PARSING_UD,
		SMS_PARSING_RD,
		SMS_PARSING_SRR,
		SMS_PARSING_MR,
		SMS_PARSING_DA,
		
		SMS_PARSING_UDH_CONCAT,
		SMS_PARSING_UDH_LRGIMG,
		SMS_PARSING_UDH_SMLIMG
	}
	
	// Stack used to store the states of the parser
	private Stack< State > m_state;
	
	// Current path
	private String [] m_path;
	private String m_pathname;
	
	// Structure of the current "raw" file
	private byte m_struct;
	
	// Storage for the data being parsed
	private StringBuilder m_buffer;

	// Storage of the extended languge preferemce (EF_ELP)
	private ELP m_extendedLanguagePreference;
	
	// ICCID
	private ICCID m_iccid;
	
	// Storage for the language preference (EF_LP)
	private LP m_languagePreference;
	
	// IMSI
	private IMSI m_imsi;
	
	// Storage for the PLMN (EF_PLMNsel, EF_FPLMN)
	private PLMN m_plmn;
	
	// Storage for SST data
	private SST m_sst;
	private int m_sstServiceNumber;
	private boolean m_sstAllocated, m_sstActivated;
	
	// Storage for the SPN data
	private SPN m_spn;
	
	// Storage for the LOCI data
	private LOCI m_loci;
	
	// Storage for the LOCIGPRS data
	private LOCIGPRS m_locigprs;
	
	// Storage for Administrative Data (AD)
	private AD m_ad;
	
	// Storage for the SIM's phase
	private PHASE m_phase;
	
	// Storage for the TON/NPI
	private Address m_address;
	
	// Storage for the phone book entries beeing parsed (EF_ADN, EF_FDN, EF_MSISDN, EF_LND)
	private Contact m_contact;
	
	// Storage for extension data (EF_EXT1, EF_EXT2, EF_EXT3, EF_EXT4)
	private Extension m_extension;
	
	// Storage for the SMS entry beeing parsed (EF_SMS)
	private SMS m_sms;
	
	// Index of the last record updated
	private int m_lastRecIndex;
}


