/*
 * This software was developed at the National Institute of Standards and Technology
 * by employees of the Federal Government in the course of their official duties.
 * Pursuant to Title 17 Section 105 of the United States Code this software is not
 * subject to copyright protection within the United States and is in the public domain.
 * The software is an experimental system. NIST assumes no responsibility whatsoever for
 * its use by any party.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS." YOU ACCEPT THE SOFTWARE "AS IS" AND ACKNOWLEDGE
 * THAT NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, INCLUDING,
 * WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY. NIST DOES NOT REPRESENT OR WARRANT THAT
 * THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT ANY
 * DEFECTS WILL BE CORRECTED. NIST DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS
 * REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED
 * TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.
 * 
 * You may freely use, copy or distribute this software.Please include appropriate
 * attribution.
 */

package com.simfill.gsm;

import com.jaccal.util.NumUtil;

import com.simfill.LogManager;
import com.simfill.Registry;

import java.lang.StringBuffer;
import java.lang.String;

// Stores the data to be stored in a record of a phone book (EF_ADN, EF_FDN, EF_MSISDN, EF_LND)
public class Contact extends LinearFixedFile
{
	// Encodings for the name field
	public enum Encoding { ENC_DEFAULT, ENC_UCS2 };
	
	public Contact( short size )
	{
		super( size );
		
		setNameEncoding( Encoding.ENC_DEFAULT );
		setName( "" );
		setAddress( new Address() );
		setCapability( ( byte )-1 );
		setExtension ( ( byte )-1 );
	}

	public void setNameEncoding( Encoding enc )
	{
		m_enc = enc;
	}
	
	public void setName( String name )
	{
		m_name = name;
	}
	
	public void setAddress( Address address )
	{
		m_address = address;
	}
	
	public void setCapability( byte capability )
	{
		m_capability = capability;
	}
	
	public void setExtension( byte extension )
	{
		m_extension = extension;
	}
	
	public String toHex()
	{
		int idx = 0;
		byte [] storage = new byte [ getSize() ];
		
		// Calculate the space available for the name in an entry on the SIM
		byte maxNamSize = ( byte )( getSize() - 14 );

		// Write the name in the appropriate encoding 
		if( m_name.length() > 0 )
		switch( m_enc )
		{
			case ENC_UCS2:

				idx = toHexUcs2( storage, idx, maxNamSize );
				break;

			case ENC_DEFAULT:
				
				idx = toHexDefault( storage, idx, maxNamSize );
				break;

			default:
				
				LogManager.getInstance().log( "warn", "Encoding '" + m_enc + "' is not supported. No name will be written for this entry.", true );
		}
		
		// Add padding in the remaining space
		while( idx < maxNamSize )
		{
			storage[ idx++ ] = -1;
		}
		
		// Check the phone number's size and truncate it if necessary
		String number = m_address.getNumber();
		byte maxNumSize = 20;
		byte numSize = number.length() < ( int )maxNumSize ? ( byte )number.length() : maxNumSize;
		
		// Give a warning if the name has been truncated
		if( numSize < number.length() )
		{
			LogManager.getInstance().log( "warn", "The maximum size of a phone number is " + maxNumSize + ". The entry '" + number + "' will be truncated. You might want to consider using an extension file for longer number. See EF_EXT*.", true );
		}
		
		// Store the number of bytes used to represent the number 
		storage[ idx++ ] = ( byte )( ( numSize >> 1 ) + ( numSize & 1 ) + 1 );
		
		// Store the type of address
		storage[ idx++ ] = m_address.getTypeOfAddress();
		
		// Store the phone number
		for( byte digit = 0; digit < numSize; digit += 2 )
		{
			byte dl = ( byte )( number.charAt( digit ) - '0' );
			byte dh = ( byte )( digit + 1 < numSize ? ( byte )number.charAt( digit + 1 ) - '0' : -1 );
			storage[ idx++ ] = ( byte )( dl + ( dh << 4 ) );
		}

		// Add padding in the remaining space
		while( idx < getSize() - 2 )
		{
			storage[ idx++ ] = -1;
		}

		// Add the optional fields
		storage[ idx++ ] = m_capability;
		storage[ idx++ ] = m_extension;
		
		return NumUtil.hex2String( storage );
	}
	
	private int toHexDefault( byte [] storage, int idx, byte maxNamSize )
	{
		// Check the name's size and truncate it if necessary
		byte [] trName = Charset.translate( m_name );
		byte namSize = trName.length < ( int )maxNamSize ? ( byte )trName.length : maxNamSize;

		// If the truncation happens on a 2-bytes special character, also remove the remaining byte
		if( trName[ namSize - 1 ] == Charset.CHARSET_ESCAPE ) --namSize;
	
		// Store the truncated name
		setName( Charset.untranslate( trName, 0, namSize ) );

		// Give a warning if the name has been truncated
		if( namSize < trName.length )
		{
			LogManager.getInstance().log( "warn", "The maximum size of a contact's name is " + maxNamSize + ". The entry '" + m_name + "' will be truncated.", true );
		}
		
		// Write the name to the buffer
		for( int trIdx = 0; trIdx < namSize; ++trIdx )
		{
			storage[ idx++ ] = trName[ trIdx ];
		}

		return idx;
	}

	private int toHexUcs2( byte [] storage, int idx, byte maxNamSize )
	{
		// Check the name's size and truncate it if necessary
		byte namBufSize = ( byte )( ( maxNamSize - 1 ) >> 1 );
		byte namSize = m_name.length() < ( int )namBufSize ? ( byte )m_name.length() : namBufSize;

		// Give a warning if the name has been truncated
		if( namSize < m_name.length() )
		{
			LogManager.getInstance().log( "warn", "The maximum size of a name in the phone book is " + namBufSize + ". The entry '" + m_name + "' will be truncated.", true );
		}
		
		// Store the truncated name
		setName( m_name.substring( 0, namSize ) );
		
		// FIXME
		storage[ idx++ ] = ( byte )0x80;

		// Store the name to the buffer
		for( char c : m_name.toCharArray() )
		{
			storage[ idx++ ] = ( byte )( ( c & 0xff00 ) >> 8 );
			storage[ idx++ ] = ( byte )( ( c & 0x00ff ) >> 0 );
		}

		return idx;
	}
	
	public String toString()
	{
		StringBuffer report = new StringBuffer();
		
		String newline = Registry.getInstance().get( "report.newline" );

		report.append( "Name: " );
		report.append( m_name );
		report.append( newline );

		report.append( "Encoding: " );
		report.append( m_enc == Encoding.ENC_UCS2 ? "UCS2" : "default 7-bit" );
		report.append( newline );

		report.append( "Address:" );
		report.append( newline );
		if( m_address.getNumber().length() > 20 )
		{
			Address a = new Address();
			a.setTON( m_address.getTON() );
			a.setNPI( m_address.getNPI() );
			a.setNumber( m_address.getNumber().substring( 0, 20 ) );
			report.append( a );
		}
		else
		{
			report.append( m_address );
		}
		
		report.append( "Capability/configuration identifier: " );
		report.append( m_capability == -1 ? "not set" : NumUtil.getUnsignedValue( m_capability ) );
		report.append( newline );
		
		report.append( "Extension: " );
		report.append( m_extension == -1 ? "not set" : NumUtil.getUnsignedValue( m_extension ) );
		report.append( newline );
		
		return report.toString();
	}
	
	// Encoding of the name entry
	private Encoding m_enc;
	
	// Name of the entry
	private String m_name;

	// Number of the entry
	private Address m_address;
	
	// Capability/configuration identifier - see EF_CCP
	private byte m_capability;
	
	// Extension record identifier - see EF_EXT?
	private byte m_extension;
}
