/*
 * This software was developed at the National Institute of Standards and Technology
 * by employees of the Federal Government in the course of their official duties.
 * Pursuant to Title 17 Section 105 of the United States Code this software is not
 * subject to copyright protection within the United States and is in the public domain.
 * The software is an experimental system. NIST assumes no responsibility whatsoever for
 * its use by any party.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS." YOU ACCEPT THE SOFTWARE "AS IS" AND ACKNOWLEDGE
 * THAT NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, INCLUDING,
 * WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY. NIST DOES NOT REPRESENT OR WARRANT THAT
 * THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT ANY
 * DEFECTS WILL BE CORRECTED. NIST DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS
 * REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED
 * TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.
 * 
 * You may freely use, copy or distribute this software.Please include appropriate
 * attribution.
 */

package com.simfill.gsm;

import com.jaccal.util.NumUtil;
import com.simfill.Registry;

import java.lang.StringBuilder;

public class LOCI extends TransparentFile
{
	// Location update status, as specified in GSM 11.11
	public static final byte STATUS_UPDATED              = 0x0;
	public static final byte STATUS_NOT_UPDATED          = 0x1;
	public static final byte STATUS_PLMN_NOT_ALLOWED     = 0x2;
	public static final byte STATUS_LOC_AREA_NOT_ALLOWED = 0x3;
	public static final byte STATUS_RESERVED             = 0x7;

	public LOCI( int size ) throws Exception
	{
		super( size );
		
		if( getSize() != 11 )
		{
			throw new Exception( "LOCI: the size of the file EF_LOCI is not standard." );
		}
		
		m_tmsi = new byte[]{ ( byte )0xff, ( byte )0xff, ( byte )0xff, ( byte )0xff };
		m_mcc = "000";
		m_mnc = "000";
		m_lac = new byte[]{ ( byte )0xff, ( byte )0xff };
		m_tmsiTime = 0;
		m_updateStatus = STATUS_RESERVED;
	}

	public void setTMSI( String tmsi ) throws Exception
	{
		if( tmsi.length() != 8 )
		{
			throw new Exception( "LOCI: the size of TMSI in EF_LOCI should be 4 bytes." );
		}

		m_tmsi = NumUtil.toStringHex( tmsi );
	}
	
	public void setMCC( String mcc ) throws Exception
	{
		if( mcc.length() != 3 )
		{
			throw new Exception( "LOCI: the length of LAI:MCC in EF_LOCI should be 3 characters." );
		}

		m_mcc = mcc;
	}

	public void setMNC( String mnc ) throws Exception
	{
		switch( mnc.length() )
		{
			case 2:
			case 3:
				
				break;
				
			default:

				throw new Exception( "LOCI: the length of LAI:MNC in EF_LOCI should be 2 or 3 characters." );
		}

		m_mnc = mnc;
	}
	
	public void setLAC( String lac ) throws Exception
	{
		if( lac.length() != 4 )
		{
			throw new Exception( "LOCI: the size of LAI:LAC in EF_LOCI should be 2 bytes." );
		}
		
		m_lac = NumUtil.toStringHex( lac );
	}

	public void setTMSITime( String time ) throws Exception
	{
		if( time.length() != 2 )
		{
			throw new Exception( "LOCI: the size of TMSI TIME in EF_LOCI should be 1 byte." );
		}
		
		m_tmsiTime = NumUtil.toStringHex( time )[ 0 ];
	}

	public void setLocationUpdateStatus( byte status )
	{
		m_updateStatus = status;
	}

	public String toHex()
	{
		byte [] storage = new byte[ getSize() ];
		int index = 0;
		
		// Store the TMSI
		for( byte b : m_tmsi ) storage[ index++ ] = b;
		
		// Store the LAI (MCC, MNC and LAC)
		String mnc = m_mnc + ( char )( 0xf + '0' );
		storage[ index++ ] = ( byte )( ( ( m_mcc.charAt( 1 ) - '0' ) << 4 ) + ( m_mcc.charAt( 0 ) - '0' ) );
		storage[ index++ ] = ( byte )( ( (   mnc.charAt( 2 ) - '0' ) << 4 ) + ( m_mcc.charAt( 2 ) - '0' ) ); 
		storage[ index++ ] = ( byte )( ( (   mnc.charAt( 1 ) - '0' ) << 4 ) + (   mnc.charAt( 0 ) - '0' ) );

		for( byte b : m_lac ) storage[ index++ ] = b;
		
		// Store the TMSI time
		storage[ index++ ] = m_tmsiTime;
		
		// Store the update status
		storage[ index++ ] = m_updateStatus;

		return NumUtil.hex2String( storage );
	}

	public String toString()
	{
		StringBuilder report = new StringBuilder();
		
		String newline = Registry.getInstance().get( "report.newline" );
		
		report.append( "TMSI: " );
		report.append( NumUtil.hex2String( m_tmsi ) );
		report.append( newline );
		
		report.append( "LAI/MCC: " );
		report.append ( m_mcc );
		report.append( newline );

		report.append( "LAI/MNC: " );
		report.append ( m_mnc );
		report.append( newline );
		
		report.append( "LAI/LAC: " );
		report.append( NumUtil.hex2String( m_lac ) );
		report.append( newline );

		report.append( "TMSI time: " );
		report.append( NumUtil.hex2String( m_tmsiTime ) );
		report.append( newline );

		report.append( "Update status: " );
		switch( m_updateStatus )
		{
			case STATUS_UPDATED:              report.append( "updated"                   ); break;
			case STATUS_NOT_UPDATED:          report.append( "not updated"               ); break;
			case STATUS_PLMN_NOT_ALLOWED:     report.append( "PLMN not allowed"          ); break;
			case STATUS_LOC_AREA_NOT_ALLOWED: report.append( "location area not allowed" ); break;
			case STATUS_RESERVED:             report.append( "reserved"                  ); break;
			default:                          report.append( "INVALID"                   );
		}
		report.append( newline );
		
		return report.toString();
	}
	
	// TMSI
	private byte [] m_tmsi;
	
	// LAI
	private String m_mcc, m_mnc;
	private byte [] m_lac;
	
	// TMSI time
	private byte m_tmsiTime;
	
	// Location update status
	private byte m_updateStatus;
}
