/*
 * This software was developed at the National Institute of Standards and Technology
 * by employees of the Federal Government in the course of their official duties.
 * Pursuant to Title 17 Section 105 of the United States Code this software is not
 * subject to copyright protection within the United States and is in the public domain.
 * The software is an experimental system. NIST assumes no responsibility whatsoever for
 * its use by any party.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS." YOU ACCEPT THE SOFTWARE "AS IS" AND ACKNOWLEDGE
 * THAT NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, INCLUDING,
 * WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY. NIST DOES NOT REPRESENT OR WARRANT THAT
 * THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT ANY
 * DEFECTS WILL BE CORRECTED. NIST DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS
 * REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED
 * TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.
 * 
 * You may freely use, copy or distribute this software.Please include appropriate
 * attribution.
 */

package com.simfill.gsm;

import com.jaccal.util.NumUtil;
import com.simfill.Registry;

import java.lang.StringBuilder;

public class LOCIGPRS extends TransparentFile
{
	// Location update status, as specified in GSM 11.11
	public static final byte STATUS_UPDATED               = 0x0;
	public static final byte STATUS_NOT_UPDATED           = 0x1;
	public static final byte STATUS_PLMN_NOT_ALLOWED      = 0x2;
	public static final byte STATUS_ROUT_AREA_NOT_ALLOWED = 0x3;
	public static final byte STATUS_RESERVED              = 0x7;

	public LOCIGPRS( int size ) throws Exception
	{
		super( size );
		
		if( getSize() != 14 )
		{
			throw new Exception( "LOCIGPRS: the size of the file EF_LOCIGPRS is not standard." );
		}
		
		m_ptmsi          = new byte[]{ ( byte )0xff, ( byte )0xff, ( byte )0xff, ( byte )0xff };
		m_ptmsiSignature = new byte[]{ ( byte )0xff, ( byte )0xff, ( byte )0xff };
		m_mcc            = "000";
		m_mnc            = "000";
		m_lac            = new byte[]{ ( byte )0xff, ( byte )0xff };
		m_updateStatus   = STATUS_RESERVED;
	}

	public void setPTMSI( String ptmsi ) throws Exception
	{
		if( ptmsi.length() != 8 )
		{
			throw new Exception( "LOCIGPRS: the size of P-TMSI in EF_GPRSLOCI should be 4 bytes." );
		}

		m_ptmsi = NumUtil.toStringHex( ptmsi );
	}
	
	public void setPTMSISignature( String signature ) throws Exception
	{
		if( signature.length() != 6 )
		{
			throw new Exception( "LOCIGPRS: the length of the P-TMSI signature in EF_LOCIGPRS should be 3 bytes." );
		}

		m_ptmsiSignature = NumUtil.toStringHex( signature );
	}

	public void setMCC( String mcc ) throws Exception
	{
		if( mcc.length() != 3 )
		{
			throw new Exception( "LOCIGPRS: the length of RAI:MCC in EF_LOCIGPRS should be 3 characters." );
		}

		m_mcc = mcc;
	}

	public void setMNC( String mnc ) throws Exception
	{
		switch( mnc.length() )
		{
			case 2:
			case 3:
				
				break;
				
			default:

				throw new Exception( "LOCIGPRS: the length of RAI:MNC in EF_LOCIGPRS should be 2 or 3 characters." );
		}

		m_mnc = mnc;
	}
	
	public void setLAC( String lac ) throws Exception
	{
		if( lac.length() != 4 )
		{
			throw new Exception( "LOCIGPRS: the size of RAI:LAC in EF_LOCIGPRS should be 2 bytes." );
		}
		
		m_lac = NumUtil.toStringHex( lac );
	}	

	public void setRAC( String rac ) throws Exception
	{
		if( rac.length() != 2 )
		{
			throw new Exception( "LOCIGPRS: the size of RAI:RAC in EF_LOCIGPRS should be 1 byte." );
		}
		
		m_rac = NumUtil.toStringHex( rac )[ 0 ];
	}
	
	public void setRoutingUpdateStatus( byte status )
	{
		m_updateStatus = status;
	}

	public String toHex()
	{
		byte [] storage = new byte[ getSize() ];
		int index = 0;
		
		// Store the P-TMSI
		for( byte b : m_ptmsi )				storage[ index++ ] = b;
		
		// Store the P-TMSI signature
		for( byte b : m_ptmsiSignature )	storage[ index++ ] = b;

		// Store the RAI (MCC, MNC, LAC and RAC)
		String mnc = m_mnc + ( char )( 0xf + '0' );
		storage[ index++ ] = ( byte )( ( ( m_mcc.charAt( 1 ) - '0' ) << 4 ) + ( m_mcc.charAt( 0 ) - '0' ) );
		storage[ index++ ] = ( byte )( ( (   mnc.charAt( 2 ) - '0' ) << 4 ) + ( m_mcc.charAt( 2 ) - '0' ) ); 
		storage[ index++ ] = ( byte )( ( (   mnc.charAt( 1 ) - '0' ) << 4 ) + (   mnc.charAt( 0 ) - '0' ) );

		for( byte b : m_lac ) storage[ index++ ] = b;
		
		storage[ index++ ] = m_rac;

		// Store the update status
		storage[ index++ ] = m_updateStatus;

		return NumUtil.hex2String( storage );
	}

	public String toString()
	{
		StringBuilder report = new StringBuilder();
		
		String newline = Registry.getInstance().get( "report.newline" );
		
		report.append( "P-TMSI: " );
		report.append( NumUtil.hex2String( m_ptmsi ) );
		report.append( newline );
		
		report.append( "P-TMSI signature: " );
		report.append ( NumUtil.hex2String( m_ptmsiSignature ) );
		report.append( newline );

		report.append( "RAI/MCC: " );
		report.append ( m_mcc );
		report.append( newline );

		report.append( "RAI/MNC: " );
		report.append ( m_mnc );
		report.append( newline );
		
		report.append( "RAI/LAC: " );
		report.append( NumUtil.hex2String( m_lac ) );
		report.append( newline );

		report.append( "RAI/RAC: " );
		report.append( NumUtil.hex2String( m_rac ) );
		report.append( newline );

		report.append( "Update status: " );
		switch( m_updateStatus )
		{
			case STATUS_UPDATED:               report.append( "updated"                   ); break;
			case STATUS_NOT_UPDATED:           report.append( "not updated"               ); break;
			case STATUS_PLMN_NOT_ALLOWED:      report.append( "PLMN not allowed"          ); break;
			case STATUS_ROUT_AREA_NOT_ALLOWED: report.append( "routing area not allowed"  ); break;
			case STATUS_RESERVED:              report.append( "reserved"                  ); break;
			default:                           report.append( "INVALID"                   );
		}
		report.append( newline );
		
		return report.toString();
	}
	
	// P-TMSI
	private byte [] m_ptmsi;
	
	// P-TMSI signature value
	private byte [] m_ptmsiSignature;
	
	// RAI
	private String m_mcc, m_mnc;
	private byte [] m_lac;
	private byte m_rac;
	
	// Routing area update status
	private byte m_updateStatus;
}
