/*
 * This software was developed at the National Institute of Standards and Technology
 * by employees of the Federal Government in the course of their official duties.
 * Pursuant to Title 17 Section 105 of the United States Code this software is not
 * subject to copyright protection within the United States and is in the public domain.
 * The software is an experimental system. NIST assumes no responsibility whatsoever for
 * its use by any party.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS." YOU ACCEPT THE SOFTWARE "AS IS" AND ACKNOWLEDGE
 * THAT NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, INCLUDING,
 * WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY. NIST DOES NOT REPRESENT OR WARRANT THAT
 * THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT ANY
 * DEFECTS WILL BE CORRECTED. NIST DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS
 * REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED
 * TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.
 * 
 * You may freely use, copy or distribute this software.Please include appropriate
 * attribution.
 */

package com.simfill.gsm;

import java.util.Vector;

//Stores the data to be stored in a record of the file EF_SMS
public abstract class SMS extends LinearFixedFile
{
	// Status of the record in EF_SMS, according to GSM 11.11
	// First byte in the record.
	public static final byte STATUS_FREESPACE    = 0;
	public static final byte STATUS_IN_READ      = 1;
	public static final byte STATUS_IN_NOT_READ  = 3;
	public static final byte STATUS_OUT_SENT     = 5;
	public static final byte STATUS_OUT_NOT_SENT = 7;
	
	protected byte m_status;
	
	// Address of the service center
	protected Address m_sca;
	
	// TP-Message-Type-Indicator (TP-MTI), according to 3GPP TS 23.040
	// Bits 0 and 1 of the first byte of a PDU.
	protected static final byte MTI_SC2MS_SMS_DELIVER        = 0x00;
	protected static final byte MTI_SC2MS_SMS_SUBMIT_REPORT  = 0x01;
	protected static final byte MTI_SC2MS_SMS_STATUS_REPORT  = 0x02;
	protected static final byte MTI_MS2SC_SMS_DELIVER_REPORT = 0x00;
	protected static final byte MTI_MS2SC_SMS_SUBMIT         = 0x01;
	protected static final byte MTI_MS2SC_SMS_COMMAND        = 0x02;
	protected static final byte MTI_RESERVED                 = 0x03;
	
	protected byte m_mti;

	// TP-User-Data-Header-Indicator (TP-UDHI), according to 3GPP TS 23.040
	// Bit 6 of the first byte of a PDU.
	public static final byte UDHI_NO_HEADER               = 0x00;
	public static final byte UDHI_HEADER_PRESENT          = 0x40;

	protected byte m_udhi;
	
	// TP-Protocol-Identifier (TP-PID), according to 3GPP TS 23.040
	// FIXME not supported
	public static final byte PID_NOT_IMPLEMENTED          = 0x00;
	
	protected byte m_pid;
	
	// TP-User-Data-Header (TP-UDH) elements, according to 3GPP TS 23.040
	protected Vector< SMSUserDataHeaderElement > m_udh;
	
	// TP-User-Data-Header-Length (TP-UDHL), according to 3GPP TS 23.040
	protected int m_udhl;

	public SMS( short size ) throws Exception
	{
		super( size );

		if( size != 176 )
		{
			throw new Exception( "SMS: the size of a record in EF_SMS is not standard." );
		}

		setStatus( STATUS_FREESPACE    );
		setSCA   ( new Address()       );
		setMTI   ( MTI_RESERVED        );
		setUDHI  ( UDHI_NO_HEADER      );
		setUDHL  ( 0 );
		setPID   ( PID_NOT_IMPLEMENTED );
/*
		serviceCenterAddress = new Address();
		senderAddress = new Address();
		encoding = ENC_DEFAULT;
		timestamp = new byte[ 7 ];
		text = new String();
*/
	}

	public void setStatus( byte status )
	{
		m_status = status;
	}
	
	public void setSCA( Address sca )
	{
		m_sca = sca;
	}
	
	protected void setMTI( byte mti )
	{
		m_mti = mti;
	}
	
	public void setUDHI( byte udhi )
	{
		m_udhi = udhi;
	}
	
	public void setUDHL( int udhl )
	{
		m_udhl = udhl;
	}
	
	public void setPID( byte pid )
	{
		m_pid = pid;
	}
	
	public void add2UDH( SMSUserDataHeaderElement e )
	{
		if( m_udh == null )
		{
			m_udhi = UDHI_HEADER_PRESENT;
			m_udh = new Vector< SMSUserDataHeaderElement >();
		}	
	
		m_udh.add( e );
		m_udhl += e.getSize();
	}
}
