/*
 * This software was developed at the National Institute of Standards and Technology
 * by employees of the Federal Government in the course of their official duties.
 * Pursuant to Title 17 Section 105 of the United States Code this software is not
 * subject to copyright protection within the United States and is in the public domain.
 * The software is an experimental system. NIST assumes no responsibility whatsoever for
 * its use by any party.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS." YOU ACCEPT THE SOFTWARE "AS IS" AND ACKNOWLEDGE
 * THAT NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, INCLUDING,
 * WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY. NIST DOES NOT REPRESENT OR WARRANT THAT
 * THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT ANY
 * DEFECTS WILL BE CORRECTED. NIST DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS
 * REGARDING THE USE OF THE SOFTWARE OR THE RESULTS THEREOF, INCLUDING BUT NOT LIMITED
 * TO THE CORRECTNESS, ACCURACY, RELIABILITY, OR USEFULNESS OF THE SOFTWARE.
 * 
 * You may freely use, copy or distribute this software.Please include appropriate
 * attribution.
 */

package com.simfill.gsm;

import com.jaccal.util.NumUtil;

import com.simfill.LogManager;
import com.simfill.Registry;

import java.lang.StringBuffer;
import java.lang.String;

// Stores the data of EF_SPN
public class SPN extends TransparentFile
{
	// Encodings for the name field
	public enum Encoding { ENC_DEFAULT, ENC_UCS2 };
	
	// States of the display condition
	public enum DisplayCondition { NOT_REQUIRED, REQUIRED };
	
	public SPN( int size ) throws Exception
	{
		super( size );
		
		if( getSize() != 17 )
		{
			throw new Exception( "SPN: the size of the file EF_SPN is not standard." );
		}

		setDisplayCondition( DisplayCondition.REQUIRED );
		setNameEncoding( Encoding.ENC_DEFAULT );
		setSPN( "" );
	}
	
	public void setDisplayCondition( DisplayCondition dcond )
	{
		m_dcond = dcond;
	}

	public void setNameEncoding( Encoding enc )
	{
		m_enc = enc;
	}
	
	public void setSPN( String spn )
	{
		m_spn = spn;
	}
	
	public String toHex()
	{
		switch( m_enc )
		{
			case ENC_UCS2:

				return toHexUcs2();

			default:
				
				return toHexDefault();
		}
	}
	
	private String toHexDefault()
	{
		byte [] storage = new byte [ getSize() ];

		byte [] trSpn = Charset.translate( m_spn );

		int length = trSpn.length;

		if( length > 16 )
		{
			LogManager.getInstance().log( "warn", "The maximum size of a service provider name is 16. The entry '" + m_spn + "' will be truncated.", true );
			length = 16;
		}
		
		// If the truncation happens on a 2-bytes special character, also remove the remaining byte
		if( trSpn[ length - 1 ] == Charset.CHARSET_ESCAPE ) --length;
		
		setSPN( Charset.untranslate( trSpn, 0, length ) );
		
		int idx = 0, trIdx = 0;
		
		storage[ idx++ ] = ( byte )( m_dcond == DisplayCondition.NOT_REQUIRED ? 0x00 : 0x01 );
		
		while( trIdx < length )
		{
			storage[ idx++ ] = trSpn[ trIdx++ ];
		}

		while( idx < storage.length )
		{
			storage[ idx++ ] = ( byte )0xff;
		}
		
		return NumUtil.hex2String( storage );
	}

	private String toHexUcs2()
	{
		byte [] storage = new byte [ getSize() ];
		
		int length = m_spn.length();

		if( length > 7 )
		{
			LogManager.getInstance().log( "warn", "The maximum size of a service provider name is 7. The entry '" + m_spn + "' will be truncated.", true );
			length = 7;
		}
		
		int idx = 0;
		
		storage[ idx++ ] = ( byte )( m_dcond == DisplayCondition.NOT_REQUIRED ? 0x00 : 0x01 );
		
		// UCS2 indicator FIXME
		storage[ idx++ ] = ( byte )0x80;
		
		for( char c : m_spn.substring( 0, length ).toCharArray() )
		{
			storage[ idx++ ] = ( byte )( ( c & 0xff00 ) >> 8 );
			storage[ idx++ ] = ( byte )( ( c & 0x00ff ) >> 0 );
		}

		while( idx < storage.length )
		{
			storage[ idx++ ] = ( byte )0xff;
		}
		
		return NumUtil.hex2String( storage );
	}

	public String toString()
	{
		StringBuffer report = new StringBuffer();
		
		String newline = Registry.getInstance().get( "report.newline" );
		
		report.append( "Display condition: " );
		report.append( m_dcond == DisplayCondition.NOT_REQUIRED ? "not required" : "required" );
		report.append( newline );

		switch( m_enc )
		{
			case ENC_DEFAULT:
				
				report.append( "Service provider name: " );
				report.append( m_spn );
				report.append( newline );

				report.append( "Encoding: 7 bits (default)" );
				report.append( newline );
				break;

			case ENC_UCS2:

				report.append( "Service provider name: " );
				report.append( m_spn.length() < 8 ? m_spn : m_spn.substring( 0, 7 ) );
				report.append( newline );

				report.append( "Encoding: UCS2" );
				report.append( newline );
				break;
		}
		
		return report.toString();
	}

	// Display condition
	private DisplayCondition m_dcond;
	
	// Encoding of the name entry
	private Encoding m_enc;
	
	// Name of the service provider
	private String m_spn;
}
